﻿/*
	DbFS.NET
    Copyright (C) 2016 SCAND Ltd, svg@scand.com

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Text;
using dbfs;
using Xunit;

public class CopyTests
{
	[Fact]
	public void CopyFolder_DestinationFolderExists()
	{
		Utils.ClearDB();

		using (DBFS dbfs = Utils.CreateDBFS())
		{
			byte[] buffer = Encoding.UTF8.GetBytes("this is content of a file");
			byte[] buffer2 = Encoding.UTF8.GetBytes("this is content of a different file");

			using (var commit = dbfs.StartCommit("", Utils.DEFAULT_USER, Utils.DEFAULT_BRANCH)) {
				commit.AddFile("/folder1/file1.txt", path => new MemoryStream(buffer));
				commit.AddFile("/folder1/file2.txt", path => new MemoryStream(buffer));
				commit.AddFile("/folder1/file4.txt", path => new MemoryStream(buffer));

				commit.AddFile("/folder2/file1.txt", path => new MemoryStream(buffer));
				commit.AddFile("/folder2/file2.txt", path => new MemoryStream(buffer2));
				commit.AddFile("/folder2/file3.txt", path => new MemoryStream(buffer));

				var result = commit.Finish();

				Assert.NotNull(result);
				Assert.Equal(9, result.Changes.Count);
			}

			using (var commit = dbfs.StartCommit("", Utils.DEFAULT_USER, Utils.DEFAULT_BRANCH)) {
				commit.CopyFolder(commit.Source("/folder1/"), "/folder2/");

				var result = commit.Finish();

				Assert.NotNull(result);
				// copied 2 files (file2.txt and file4.txt)
				// /folder2/file1.txt isn't overwritten by /folder1/file1.txt because they're identical
				Assert.Equal(2, result.Files.Count);
			}

			{
				var list = dbfs.GetFoldersList(Utils.DEFAULT_BRANCH, "/")
					.Select(e => e.Path);

				Assert.Equal(2, list.Count());
				Assert.Contains<string>("/folder1/", list);
				Assert.Contains<string>("/folder2/", list);
			}

			{
				var list = dbfs.GetFilesList(Utils.DEFAULT_BRANCH, "/folder2/")
					.Select(e => e.Path);

				Assert.Equal(4, list.Count());
				Assert.Contains<string>("/folder2/file1.txt", list);
				Assert.Contains<string>("/folder2/file2.txt", list);
				Assert.Contains<string>("/folder2/file3.txt", list);
				Assert.Contains<string>("/folder2/file4.txt", list);
			}
		}
	}

	[Fact]
	public void CopyFile_CopyRemoveCopy_Issue55()
	{
		Utils.ClearDB();

		using (DBFS dbfs = Utils.CreateDBFS())
		{
			byte[] buffer = Encoding.UTF8.GetBytes("this is content of a file");
			byte[] buffer2 = Encoding.UTF8.GetBytes("this is content of a different file");

			using (var commit = dbfs.StartCommit("", Utils.DEFAULT_USER, Utils.DEFAULT_BRANCH)) {
				commit.AddFile("/file1.txt", path => new MemoryStream(buffer));
				commit.AddFile("/file2.txt", path => new MemoryStream(buffer));
				commit.AddFile("/file4.txt", path => new MemoryStream(buffer));

				commit.AddFile("/folder1/file1.txt", path => new MemoryStream(buffer));
				commit.AddFile("/folder1/file2.txt", path => new MemoryStream(buffer2));
				commit.AddFile("/folder1/file3.txt", path => new MemoryStream(buffer));

				var result = commit.Finish();

				Assert.NotNull(result);
				Assert.Equal(8, result.Changes.Count);
			}

			using (var commit = dbfs.StartCommit("", Utils.DEFAULT_USER, Utils.DEFAULT_BRANCH)) {
				commit.CopyFile(commit.Source("/file4.txt"), "/folder1/file4.txt");

				var result = commit.Finish();

				Assert.NotNull(result);
				Assert.Equal(1, result.Changes.Count);
			}

			{
				var list = dbfs.GetFilesList(Utils.DEFAULT_BRANCH, "/folder1/")
					.Select(e => e.Path);

				Assert.Equal(4, list.Count());
				Assert.Contains<string>("/folder1/file4.txt", list);
			}

			using (var commit = dbfs.StartCommit("", Utils.DEFAULT_USER, Utils.DEFAULT_BRANCH)) {
				commit.RemoveFile("/folder1/file4.txt");

				var result = commit.Finish();

				Assert.NotNull(result);
				Assert.Equal(1, result.Changes.Count);
			}

			using (var commit = dbfs.StartCommit("", Utils.DEFAULT_USER, Utils.DEFAULT_BRANCH)) {
				commit.CopyFile(commit.Source("/file4.txt"), "/folder1/file4.txt");

				var result = commit.Finish();

				Assert.NotNull(result);
				Assert.Equal(1, result.Files.Count);
			}

			{
				var list = dbfs.GetFilesList(Utils.DEFAULT_BRANCH, "/folder1/")
					.Select(e => e.Path);

				Assert.Equal(4, list.Count());
				Assert.Contains<string>("/folder1/file4.txt", list);
			}
		}
	}

	[Fact]
	public void CopyFolder_CopyRemoveCopy_Issue55()
	{
		Utils.ClearDB();

		using (DBFS dbfs = Utils.CreateDBFS())
		{
			byte[] buffer = Encoding.UTF8.GetBytes("this is content of a file");

			using (var commit = dbfs.StartCommit("", Utils.DEFAULT_USER, Utils.DEFAULT_BRANCH)) {
				commit.AddFile("/folder1/file1.txt", path => new MemoryStream(buffer));
				commit.AddFile("/folder2/file1.txt", path => new MemoryStream(buffer));

				var result = commit.Finish();

				Assert.NotNull(result);
				Assert.Equal(5, result.Changes.Count);
			}

			using (var commit = dbfs.StartCommit("", Utils.DEFAULT_USER, Utils.DEFAULT_BRANCH)) {
				commit.CopyFolder(commit.Source("/folder2/"), "/folder1/folder2/");

				var result = commit.Finish();

				Assert.NotNull(result);
				Assert.Equal(2, result.Changes.Count);
			}

			{
				var list = dbfs.GetFoldersList(Utils.DEFAULT_BRANCH, "/folder1/")
					.Select(e => e.Path);

				Assert.Equal(1, list.Count());
				Assert.Contains<string>("/folder1/folder2/", list);
			}

			using (var commit = dbfs.StartCommit("", Utils.DEFAULT_USER, Utils.DEFAULT_BRANCH)) {
				commit.RemoveFolder("/folder1/folder2/");

				var result = commit.Finish();

				Assert.NotNull(result);
				Assert.Equal(2, result.Changes.Count);
			}

			{
				var list = dbfs.GetFoldersList(Utils.DEFAULT_BRANCH, "/folder1/")
					.Select(e => e.Path);

				Assert.Equal(0, list.Count());
			}

			using (var commit = dbfs.StartCommit("", Utils.DEFAULT_USER, Utils.DEFAULT_BRANCH)) {
				commit.CopyFolder(commit.Source("/folder2/"), "/folder1/folder2/");

				var result = commit.Finish();

				Assert.NotNull(result);
				Assert.Equal(2, result.Changes.Count);
			}

			{
				var list = dbfs.GetFoldersList(Utils.DEFAULT_BRANCH, "/folder1/")
					.Select(e => e.Path);

				Assert.Equal(1, list.Count());
				Assert.Contains<string>("/folder1/folder2/", list);
			}
		}
	}
}
