﻿/*
	DbFS.NET
    Copyright (C) 2016 SCAND Ltd, svg@scand.com

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Text;
using dbfs;
using Xunit;

public class MoveTests
{
	[Fact]
	public void MoveFile()
	{
		Utils.ClearDB();

		using (DBFS dbfs = Utils.CreateDBFS())
		{
			byte[] buffer = Encoding.UTF8.GetBytes("this is content of a file");

			using (var commit = dbfs.StartCommit("", Utils.DEFAULT_USER, Utils.DEFAULT_BRANCH)) {
				commit.AddFile("/file1.txt", p => new MemoryStream(buffer));
				commit.AddFolder("/folder1/");

				var result = commit.Finish();

				Assert.NotNull(result);
				Assert.Equal(3, result.Changes.Count);
			}

			{
				var list = dbfs.GetFoldersList(Utils.DEFAULT_BRANCH, "/")
					.Select(e => e.Path);

				Assert.Equal(1, list.Count());
				Assert.Contains<string>("/folder1/", list);


				list = dbfs.GetFilesList(Utils.DEFAULT_BRANCH, "/")
					.Select(e => e.Path);

				Assert.Equal(1, list.Count());
				Assert.Contains<string>("/file1.txt", list);
			}

			{
				var result = dbfs.MoveFile("/file1.txt", "/folder1/file1.txt", Utils.DEFAULT_USER, Utils.DEFAULT_BRANCH, "");

				Assert.NotNull(result);
				Assert.Equal(2, result.Changes.Count);
			}

			{
				var list = dbfs.GetFilesList(Utils.DEFAULT_BRANCH, "/")
					.Select(e => e.Path);

				Assert.Equal(0, list.Count());


				list = dbfs.GetFilesList(Utils.DEFAULT_BRANCH, "/folder1/")
					.Select(e => e.Path);

				Assert.Equal(1, list.Count());
				Assert.Contains<string>("/folder1/file1.txt", list);
			}
		}
	}

	[Fact]
	public void MoveFile_CreateMoved_Issue66()
	{
		MoveFile();

		using (DBFS dbfs = Utils.CreateDBFS())
		{
			byte[] buffer = Encoding.UTF8.GetBytes("this is content of a file");

			using (var commit = dbfs.StartCommit("", Utils.DEFAULT_USER, Utils.DEFAULT_BRANCH)) {
				commit.AddFile("/file1.txt", p => new MemoryStream(buffer));

				var result = commit.Finish();

				Assert.NotNull(result);
				Assert.Equal(1, result.Changes.Count);
			}

			{
				var list = dbfs.GetFilesList(Utils.DEFAULT_BRANCH, "/")
					.Select(e => e.Path);

				Assert.Equal(1, list.Count());
				Assert.Contains<string>("/file1.txt", list);
			}
		}
	}

	[Fact]
	public void MoveFile_BackAndForth_Issue66()
	{
		MoveFile();

		using (DBFS dbfs = Utils.CreateDBFS())
		{
			{
				var result = dbfs.MoveFile("/folder1/file1.txt", "/file1.txt", Utils.DEFAULT_USER, Utils.DEFAULT_BRANCH, "");

				Assert.NotNull(result);
				Assert.Equal(2, result.Changes.Count);
			}

			{
				var list = dbfs.GetFilesList(Utils.DEFAULT_BRANCH, "/")
					.Select(e => e.Path);

				Assert.Equal(1, list.Count());
				Assert.Contains<string>("/file1.txt", list);


				list = dbfs.GetFilesList(Utils.DEFAULT_BRANCH, "/folder1/")
					.Select(e => e.Path);

				Assert.Equal(0, list.Count());
			}
		}
	}

	// -------------------------------------------------------------------------------------------------------------

	[Fact]
	public void MoveFolder_DestinationFolderExists()
	{
		Utils.ClearDB();

		using (DBFS dbfs = Utils.CreateDBFS())
		{
			byte[] buffer = Encoding.UTF8.GetBytes("this is content of a file");

			using (var commit = dbfs.StartCommit("", Utils.DEFAULT_USER, Utils.DEFAULT_BRANCH)) {
				commit.AddFile("/folder1/file1.txt", path => new MemoryStream(buffer));
				commit.AddFile("/folder1/file4.txt", path => new MemoryStream(buffer));

				commit.AddFile("/folder2/file1.txt", path => new MemoryStream(buffer));
				commit.AddFile("/folder2/file2.txt", path => new MemoryStream(buffer));
				commit.AddFile("/folder2/file3.txt", path => new MemoryStream(buffer));

				var result = commit.Finish();

				Assert.NotNull(result);
				Assert.Equal(8, result.Changes.Count);
			}

			using (var commit = dbfs.StartCommit("", Utils.DEFAULT_USER, Utils.DEFAULT_BRANCH)) {
				commit.RenameFolder("/folder1/", "/folder2/");

				var result = commit.Finish();

				Assert.NotNull(result);
				// remove 2 files and 1 folder + add 1 file (file4.txt)
				// /folder2/file1.txt isn't overwritten by /folder1/file1.txt because they're identical
				Assert.Equal(4, result.Changes.Count);
			}

			{
				var list = dbfs.GetFoldersList(Utils.DEFAULT_BRANCH, "/")
					.Select(e => e.Path);

				Assert.Equal(1, list.Count());
				Assert.Contains<string>("/folder2/", list);
			}

			{
				var list = dbfs.GetFilesList(Utils.DEFAULT_BRANCH, "/folder2/")
					.Select(e => e.Path);

				Assert.Equal(4, list.Count());
				Assert.Contains<string>("/folder2/file1.txt", list);
				Assert.Contains<string>("/folder2/file2.txt", list);
				Assert.Contains<string>("/folder2/file3.txt", list);
				Assert.Contains<string>("/folder2/file4.txt", list);
			}
		}
	}

	[Fact]
	public void MoveFolder()
	{
		Utils.ClearDB();

		using (DBFS dbfs = Utils.CreateDBFS())
		{
			using (var commit = dbfs.StartCommit("", Utils.DEFAULT_USER, Utils.DEFAULT_BRANCH)) {
				commit.AddFolder("/folder1/");
				commit.AddFolder("/folder2/");

				var result = commit.Finish();

				Assert.NotNull(result);
				Assert.Equal(3, result.Changes.Count);
			}

			{
				var list = dbfs.GetFoldersList(Utils.DEFAULT_BRANCH, "/")
					.Select(e => e.Path);

				Assert.Equal(2, list.Count());
				Assert.Contains<string>("/folder1/", list);
				Assert.Contains<string>("/folder2/", list);
			}

			using (var commit = dbfs.StartCommit("", Utils.DEFAULT_USER, Utils.DEFAULT_BRANCH)) {
				commit.MoveFolderTo("/folder2/", "/folder1/");

				var result = commit.Finish();

				Assert.NotNull(result);
				Assert.Equal(2, result.Changes.Count);
			}

			{
				var list = dbfs.GetFoldersList(Utils.DEFAULT_BRANCH, "/")
					.Select(e => e.Path);

				Assert.Equal(1, list.Count());
				Assert.Contains<string>("/folder1/", list);


				list = dbfs.GetFoldersList(Utils.DEFAULT_BRANCH, "/folder1/")
					.Select(e => e.Path);

				Assert.Equal(1, list.Count());
				Assert.Contains<string>("/folder1/folder2/", list);
			}
		}
	}

	[Fact]
	public void MoveFolder_CreateMoved_Issue66()
	{
		MoveFolder();

		using (DBFS dbfs = Utils.CreateDBFS())
		{
			using (var commit = dbfs.StartCommit("", Utils.DEFAULT_USER, Utils.DEFAULT_BRANCH)) {
				commit.AddFolder("/folder2/");

				var result = commit.Finish();

				Assert.NotNull(result);
				Assert.Equal(1, result.Changes.Count);
			}

			{
				var list = dbfs.GetFoldersList(Utils.DEFAULT_BRANCH, "/")
					.Select(e => e.Path);

				Assert.Equal(2, list.Count());
				Assert.Contains<string>("/folder1/", list);
				Assert.Contains<string>("/folder2/", list);
			}
		}
	}

	[Fact]
	public void MoveFolder_BackAndForth_Issue66()
	{
		MoveFolder();

		using (DBFS dbfs = Utils.CreateDBFS())
		{
			using (var commit = dbfs.StartCommit("", Utils.DEFAULT_USER, Utils.DEFAULT_BRANCH)) {
				commit.MoveFolderTo("/folder1/folder2/", "/");

				var result = commit.Finish();

				Assert.NotNull(result);
				Assert.Equal(2, result.Changes.Count);
			}

			{
				var list = dbfs.GetFoldersList(Utils.DEFAULT_BRANCH, "/")
					.Select(e => e.Path);

				Assert.Equal(2, list.Count());
				Assert.Contains<string>("/folder1/", list);
				Assert.Contains<string>("/folder2/", list);


				list = dbfs.GetFoldersList(Utils.DEFAULT_BRANCH, "/folder1/")
					.Select(e => e.Path);

				Assert.Equal(0, list.Count());
			}
		}
	}
}
