/*
	DbFS.NET
    Copyright (C) 2016 SCAND Ltd, svg@scand.com

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using dbfs.repository;
using dbfs.utils;
using dbfs.exceptions;

namespace dbfs
{
	public class DBFS : IDisposable
	{
		public DBFS(string configName = null)
		{
			_config = configName ?? "dbfs";
			_db = DB();
		}

		private string _config;

		private Repository _db;

		#region branches

		public Branch GetBranch(int id)
		{
			return new Branch(_db.Branch.Get(id));
		}

		public Branch GetBranch(string name)
		{
			return new Branch(_db.Branch.Get(name));
		}

		public IEnumerable<Branch> ListBranches()
		{
			return _db.Branch.List().Select(e => new Branch(e));
		}

		public bool BranchExists(string name)
		{
			return _db.Branch.Exists(name);
		}

		public Branch CreateBranch(string name, string user_name, string parent_branch_name)
		{
			var user = _db.User.Get(user_name);
			var branch = _db.Branch.Create(name, user.id, parent_branch_name);

			return new Branch(branch);
		}

		public void RemoveBranch(string name)
		{
			this._db.Branch.Remove(_db.Branch.Get(name).id);
		}

		#endregion

		#region commits

		public Commit GetCommit(int commit_id)
		{
			return new Commit(_db.Commit.Get(commit_id), _db.Branch.GetByCommitId(commit_id));
		}

		public List<File> GetFilesInCommit(int commit_id)
		{
			return this.GetCommit(commit_id).Files;
		}

		public CreateCommit StartCommit(string branch_name)
		{
			int timeout = Config.Get(_config, true).GetText("commit/branch-lock-timeout").ToInt();

			if (timeout <= 0)
				timeout = 3600; // 1 hour by default

			var branch = _db.Branch.Get(branch_name);

			return new CreateCommit(this, timeout, branch.id);
		}

		public CreateCommit StartCommit(string message, string user, string branch)
		{
			return StartCommit(branch).SetMessage(message).SetUser(user);
		}

		#endregion

		#region log count

		public int GetLogCount(string branch_name)
		{
			return _db.Commit.Count(_db.Branch.Get(branch_name).id);
		}

		private int GetLogCount(string path, string branch_name)
		{
			return _db.Commit.Count(_db.Branch.Get(branch_name).id, path);
		}

		public int GetLogCountForFile(string path, string branch_name)
		{
			path.ThrowIfNull("path");

			return GetLogCount(path, branch_name);
		}

		public int GetLogCountForFolder(string directory, string branch_name)
		{
			directory.ThrowIfNull("directory");

			return GetLogCount(directory.EnsureFolder(), branch_name);
		}

		#endregion

		#region log

		public IEnumerable<Commit> GetLog(string branch_name, int start = 0, int max_count = 0)
		{
			var branch = _db.Branch.Get(branch_name);

			return _db.Commit.List(branch.id, start, max_count).Wrap(branch);
		}

		private IEnumerable<Commit> GetLog(string path, string branch_name,
			int start = 0, int max_count = 0)
		{
			var branch = _db.Branch.Get(branch_name);

			return _db.Commit.List(path, branch.id, start, max_count).Wrap(branch);
		}

		public IEnumerable<Commit> GetLogForFile(string path, string branch_name,
			int start = 0, int max_count = 0)
		{
			path.ThrowIfNull("path");

			return GetLog(path, branch_name, start, max_count);
		}

		public IEnumerable<Commit> GetLogForFolder(string directory, string branch_name,
			int start = 0, int max_count = 0)
		{
			directory.ThrowIfNull("directory");

			return GetLog(directory.EnsureFolder(), branch_name, start, max_count);
		}

		#endregion

		#region folders

		public IEnumerable<Folder> GetFoldersList(string branch_name, string directory)
		{
			directory.ThrowIfNull("directory");

			var branch = _db.Branch.Get(branch_name);

			return _db.Revision.ListOptimized(branch, directory.EnsureFolder(), 0, true, true)
				.Select(e => new Folder(e));
		}

		public IEnumerable<Folder> GetFoldersList(int commit_id, string directory)
		{
			directory.ThrowIfNull("directory");

			var branch = _db.Branch.GetByCommitId(commit_id);

			return _db.Revision.ListOptimized(branch, directory.EnsureFolder(), commit_id, true, true)
				.Select(e => new Folder(e));
		}

		public Folder GetFolder(string branch_name, string directory)
		{
			directory.ThrowIfNull("directory");

			var branch = _db.Branch.Get(branch_name);
			var revision = _db.Revision.TryGetOptimized(branch, directory);

			if (revision == null || revision.IsRemoved())
				throw new RevisionNotFoundException(directory, branch.latest_commit_id);

			return new Folder(revision);
		}

		public Folder GetFolder(int commit_id, string directory)
		{
			directory.ThrowIfNull("directory");

			var branch = _db.Branch.GetByCommitId(commit_id);
			var revision = _db.Revision.TryGetOptimized(branch, directory, commit_id);

			if (revision == null || revision.IsRemoved())
				throw new RevisionNotFoundException(directory, commit_id);

			return new Folder(revision);
		}

		public bool FolderExists(string branch_name, string path)
		{
			path.ThrowIfNull("path");

			var branch = _db.Branch.Get(branch_name);

			return _db.Revision.Exists(branch.id, path.EnsureFolder());
		}

		public bool FolderExists(int commit_id, string path)
		{
			path.ThrowIfNull("path");

			var branch = _db.Branch.GetByCommitId(commit_id);

			return _db.Revision.Exists(branch.id, path.EnsureFolder(), commit_id);
		}

		#endregion

		#region files and folders

		public IEnumerable<BaseFileEntity> GetAllNested(string branch_name, string directory)
		{
			directory.ThrowIfNull("directory");

			var branch = this._db.Branch.Get(branch_name);

			return _db.Revision.AllNestedOptimized(branch, directory.EnsureFolder())
				.Select(BaseFileEntityFactory.Create);
		}

		public IEnumerable<BaseFileEntity> GetAllNested(int commit_id, string directory)
		{
			directory.ThrowIfNull("directory");

			var branch = _db.Branch.GetByCommitId(commit_id);

			return _db.Revision.AllNestedOptimized(branch, directory.EnsureFolder(), commit_id)
				.Select(BaseFileEntityFactory.Create);
		}

		#endregion

		#region files

		public IEnumerable<File> GetFilesList(string branch_name, string directory)
		{
			directory.ThrowIfNull("directory");

			var branch = _db.Branch.Get(branch_name);

			return _db.Revision.ListOptimized(branch, directory.EnsureFolder(), 0, false, true)
				.Select(e => new File(e));
		}

		public IEnumerable<File> GetFilesList(int commit_id, string directory)
		{
			directory.ThrowIfNull("directory");

			var branch = _db.Branch.GetByCommitId(commit_id);

			return _db.Revision.ListOptimized(branch, directory.EnsureFolder(), commit_id, false, true)
				.Select(e => new File(e));
		}

		/// <exception cref="RevisionNotFoundException"></exception>
		public File GetFile(string branch_name, string path)
		{
			path.ThrowIfNull("path");

			var branch = _db.Branch.Get(branch_name);
			var revision = _db.Revision.TryGetOptimized(branch, path);

			if (revision == null || revision.IsRemoved())
				throw new RevisionNotFoundException(path, branch.latest_commit_id);

			return new File(revision);
		}

		/// <exception cref="RevisionNotFoundException"></exception>
		public File GetFile(int commit_id, string path)
		{
			path.ThrowIfNull("path");

			var branch = _db.Branch.GetByCommitId(commit_id);
			var revision = _db.Revision.TryGetOptimized(branch, path, commit_id);

			if (revision == null || revision.IsRemoved())
				throw new RevisionNotFoundException(path, commit_id);

			return new File(revision);
		}

		public bool FileExists(string branch_name, string path)
		{
			path.ThrowIfNull("path");

			var branch = _db.Branch.Get(branch_name);

			return _db.Revision.Exists(branch.id, path);
		}

		public bool FileExists(int commit_id, string path)
		{
			path.ThrowIfNull("path");

			var branch = this._db.Branch.GetByCommitId(commit_id);

			return _db.Revision.Exists(branch.id, path, commit_id);
		}

		public Stream GetContent(File file)
		{
			if (file.BlobID == 0) throw new ArgumentException();

			return _db.Blob.GetContent(file.BlobID);
		}

		public Stream GetContent(string hash)
		{
			if (hash == null || hash.Length != 40) throw new ArgumentException();

			return _db.Blob.GetContent(hash);
		}

		#endregion

		#region copy and move methods

		public Commit CopyFile(string from, int commit_id, string to, string user_name, string branch, string message)
		{
			using (var commit = StartCommit(message, user_name, branch))
			{
				commit.CopyFile(commit.Source(from, commit_id), to);

				return commit.Finish();
			}
		}

		public Commit CopyFolder(string from, int commit_id, string to, string user_name, string branch, string message)
		{
			using (var commit = StartCommit(message, user_name, branch))
			{
				commit.CopyFolder(commit.Source(from, commit_id), to);

				return commit.Finish();
			}
		}

		/// <summary>
		/// If `to` is a folder path (e.g. ends with `/`) file will be moved into that folder.
		/// </summary>
		public Commit MoveFile(string file, string to, string user_name, string branch, string message)
		{
			using (var commit = StartCommit(message, user_name, branch))
			{
				if (to.IsFolder())
					commit.MoveFileTo(file, to);
				else
					commit.RenameFile(file, to);

				return commit.Finish();
			}
		}

		public Commit MoveFolder(string from, string to, string user_name, string branch, string message)
		{
			using (var commit = StartCommit(message, user_name, branch))
			{
				if (to.IsFolder())
					commit.MoveFolderTo(from, to);
				else
					commit.RenameFolder(from, to);

				return commit.Finish();
			}
		}

		#endregion

		internal Repository DB()
		{
			return new Repository(_config);
		}

		internal void _reload()
		{
			_db.Reload();
		}

		#region IDisposable implementation

		public void Dispose()
		{
			_db.Dispose();
		}

		#endregion
	}
}
