﻿/*
	DbFS.NET
    Copyright (C) 2016 SCAND Ltd, svg@scand.com

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace dbfs.utils
{
	public class FileUrl : IEquatable<FileUrl>, IConvertible, ICloneable
	{
		public static readonly string FILE_PROTOCOL = "file";

		public FileUrl() {}

		public FileUrl(string url)
		{
			InitFromString(url);
		}

		public FileUrl(FileUrl baseDir, string relativePath)
		{
			InitFromBase(baseDir, relativePath);
		}

		public string Protocol { get; set; }
		public string UserInfo { get; set; }
		public string Host { get; set; }
		public string QueryString { get; set; }
		public string Anchor { get; set; }

		public string Directory
		{
			get { return _directory ?? ""; }
			set { _directory = FileUtils.NormalizePath(value); }
		}

		public string FileName
		{
			get { return _filename ?? ""; }
			set
			{
				string full_name = FileUtils.CombinePath(Directory, value);
				_directory = FileUtils.GetDirectoryName(full_name);
				_filename = FileUtils.GetFileName(full_name);
			}
		}

		public void SetLocalPath(string path)
		{
			Directory = "";
			FileName = path;
		}

		public string Extension
		{
			get
			{
				int pos = FileName.LastIndexOf('.');
				return pos > 0 ? FileName.Substring(pos) : "";
			}
		}

		public string FileNameWithoutExtension
		{
			get
			{
				int pos = FileName.LastIndexOf('.');
				return pos > 0 ? FileName.Substring(0, pos) : FileName;
			}
		}

		public virtual string LocalPath
		{
			get
			{
				if (string.IsNullOrEmpty(Directory))
					return FileName;

				return string.Concat(Directory, "/", FileName);
			}
		}

		public virtual string RemoteDirectory
		{
			get
			{
				return string.Concat(UrlPrefix, Directory);
			}
		}

		public virtual string RemotePath
		{
			get
			{
				return string.Concat(UrlPrefix, LocalPath);
			}
		}

		public virtual string Url
		{
			get
			{
				//TODO: Uri.EscapeUriString() might be needed
				return string.Concat(RemotePath, UrlSuffix);
			}
		}

		public Dictionary<string, string> Params
		{
			get
			{
				Dictionary<string, string> all = new Dictionary<string, string>();

				if (string.IsNullOrEmpty(QueryString))
					return all;

				string[] pairs = (QueryString ?? "").Split('&');

				foreach (string pair in pairs)
				{
					int pos = pair.IndexOf('=');

					if (pos >= 0)
						all[Uri.UnescapeDataString(pair.Substring(0, pos))] = Uri.UnescapeDataString(pair.Substring(pos + 1));
					else
						all[Uri.UnescapeDataString(pair)] = "";
				}

				return all;
			}
		}

		public virtual bool IsEmpty
		{
			get
			{
				return string.IsNullOrEmpty(Protocol)
					&& string.IsNullOrEmpty(UserInfo)
					&& string.IsNullOrEmpty(Host)
					&& string.IsNullOrEmpty(Directory)
					&& string.IsNullOrEmpty(FileName)
					&& string.IsNullOrEmpty(QueryString)
					&& string.IsNullOrEmpty(Anchor)
				;
			}
		}

		public bool IsLocal
		{
			get
			{
				return string.IsNullOrEmpty(Protocol) || Protocol == FILE_PROTOCOL;
			}
		}

		public FileUrl Parent
		{
			get
			{
				FileUrl parent = NewInstance();
				parent.InitFromString(string.Concat(RemoteDirectory, UrlSuffix));
				return parent;
			}
		}

		public virtual void InitFrom(FileUrl src)
		{
			Protocol = src.Protocol;
			UserInfo = src.UserInfo;
			Host = src.Host;
			Directory = src.Directory;
			FileName = src.FileName;
			QueryString = src.QueryString;
			Anchor = src.Anchor;
		}

		public virtual void InitFromBase(FileUrl baseDir, string relativePath)
		{
			if (FileUtils.IsAbsolutePath(relativePath))
				InitFromString(relativePath);
			else
				InitFromString(string.Concat(baseDir.RemotePath, "/", relativePath, baseDir.UrlSuffix));
		}

		public virtual void InitFromString(string url)
		{
			if (string.IsNullOrEmpty(url))
			{
				Protocol = null;
				UserInfo = null;
				Host = null;
				QueryString = null;
				Anchor = null;
				Directory = null;
				FileName = null;
			}
			else if (FileUtils.IsUrl(url))
			{
				Uri uri = new Uri(url);

				Protocol = uri.Scheme;
				UserInfo = uri.UserInfo;
				Host = uri.Host + (uri.IsDefaultPort ? "" : ":" + uri.Port);
				QueryString = uri.Query.TrimStart('?');
				Anchor = uri.Fragment.TrimStart('#');
				SetLocalPath(string.Join("", uri.Segments.Select(s => Uri.UnescapeDataString(s))).TrimStart('/'));

				// check if Uri failed to parse and included query string into file name
				int pos = FileName.IndexOf('?');

				if (QueryString.Length == 0 && pos >= 0)
				{
					QueryString = FileName.Substring(pos + 1);
					FileName = FileName.Substring(0, pos);
				}
			}
			else
			{
				Protocol = FILE_PROTOCOL;
				UserInfo = null;
				Host = null;
				QueryString = null;
				Anchor = null;
				SetLocalPath(FileUtils.GetFullPath(url));
			}
		}

		public virtual string UrlPrefix
		{
			get
			{
				if (string.IsNullOrEmpty(Protocol))
					return "";

				StringBuilder prefix = new StringBuilder();

				prefix.Append(Protocol);
				prefix.Append("://");

				if (!string.IsNullOrEmpty(UserInfo))
				{
					prefix.Append(UserInfo);
					prefix.Append("@");
				}

				prefix.Append(Host);
				prefix.Append('/');

				return prefix.ToString();
			}
		}

		public virtual string UrlSuffix
		{
			get
			{
				StringBuilder suffix = new StringBuilder();

				if (!string.IsNullOrEmpty(QueryString))
				{
					suffix.Append('?');
					suffix.Append(QueryString);
				}

				if (!string.IsNullOrEmpty(Anchor))
				{
					suffix.Append('#');
					suffix.Append(Anchor);
				}

				return suffix.ToString();
			}
		}

		public override string ToString()
		{
			return Url;
		}

		private string _directory;
		private string _filename;

		#region IEquatable<ResourceUrl>

		public static bool operator ==(FileUrl left, FileUrl right)
		{
			if (ReferenceEquals(left, right))
				return true;

			if (ReferenceEquals(left, null))
				return false;

			return left.Equals(right);
		}

		public static bool operator !=(FileUrl left, FileUrl right)
		{
			return !(left == right);
		}

		public virtual bool Equals(FileUrl other)
		{
			if (object.ReferenceEquals(other, null))
				return false;

			return Url == other.Url;
		}

		public override bool Equals(object obj)
		{
			return Equals(obj as FileUrl);
		}

		public override int GetHashCode()
		{
			return Url.GetHashCode();
		}

		#endregion

		#region IConvertible

		string IConvertible.ToString(IFormatProvider provider)
		{
			return Url;
		}

		object IConvertible.ToType(Type conversionType, IFormatProvider provider)
		{
			if (typeof(string).Equals(conversionType))
				return Url;

			throw new InvalidCastException();
		}

		TypeCode IConvertible.GetTypeCode() { return TypeCode.Object; }
		bool IConvertible.ToBoolean(IFormatProvider provider) { throw new InvalidCastException(); }
		byte IConvertible.ToByte(IFormatProvider provider) { throw new InvalidCastException(); }
		char IConvertible.ToChar(IFormatProvider provider) { throw new InvalidCastException(); }
		DateTime IConvertible.ToDateTime(IFormatProvider provider) { throw new InvalidCastException(); }
		decimal IConvertible.ToDecimal(IFormatProvider provider) { throw new InvalidCastException(); }
		double IConvertible.ToDouble(IFormatProvider provider) { throw new InvalidCastException(); }
		short IConvertible.ToInt16(IFormatProvider provider) { throw new InvalidCastException(); }
		int IConvertible.ToInt32(IFormatProvider provider) { throw new InvalidCastException(); }
		long IConvertible.ToInt64(IFormatProvider provider) { throw new InvalidCastException(); }
		sbyte IConvertible.ToSByte(IFormatProvider provider) { throw new InvalidCastException(); }
		float IConvertible.ToSingle(IFormatProvider provider) { throw new InvalidCastException(); }
		ushort IConvertible.ToUInt16(IFormatProvider provider) { throw new InvalidCastException(); }
		uint IConvertible.ToUInt32(IFormatProvider provider) { throw new InvalidCastException(); }
		ulong IConvertible.ToUInt64(IFormatProvider provider) { throw new InvalidCastException(); }

		#endregion

		#region IClonable

		public FileUrl Copy()
		{
			FileUrl url = NewInstance();
			url.InitFrom(this);
			return url;
		}

		object ICloneable.Clone()
		{
			return Copy();
		}

		protected FileUrl NewInstance()
		{
			return (FileUrl)Activator.CreateInstance(GetType());
		}

		#endregion
	}
}
